/*****************************************************************************
 * Copyright © 2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
 ****************************************************************************/

// NOTE: Requires Service Architecture runtime and will not function standalone

@BaseScript(EMIBaseScript)
import groovy.transform.BaseScript

import java.text.SimpleDateFormat
import java.io.File
import groovy.json.JsonSlurper
import groovy.io.FileType

def pgmName = "logCentralize"

def cli = new CliBuilder(
	usage:  "${pgmName} [-fxvdh]",
	header: "Collects logs from Local Landing Zones into the Service\n" +
			"Architecture Landing Zone.\n" +
			"Options:",
	footer: "Exit status:\n" +
			"$RC_SUCCESS  if OK\n" +
			"$RC_WARNING  if minor issues are encountered\n" +
			"$RC_GENERAL_ERROR  if an error is encountered\n" +
			"$RC_INVALID_USAGE for invalid usage"
)
cli.with {
	f(longOpt:'force', 'Replace existing kit configurations', required: false)
	x(longOpt:'noexit', 'Do not exit the JVM on completion', required: false)
	v(longOpt:'verbose', 'Enable verbose output', required: false)
	d(longOpt:'debug', 'Enable debug output', required: false)
	h(longOpt:'help', 'Show usage information', required: false)
}

// Parse given arguments
OptionAccessor opts = cli.parse(args)
if (opts.h) {
	cli.usage()
	return 0
}
if (opts.v) {
	verbose = true
}
if (opts.d) {
	setLogLevel LogLevel.DEBUG
}
if (opts.x) { exitJVM = false }
if (opts.f) { force = true }

msg.info "Service Architecture Log Centralization - ${new Date()}"

File metricsFile = canonicalFile(metricsDir, "log_centralize.metrics")

// Results (not yet implemented)
Map results = [successes:[], warnings:[], errors:[]]

// Metrics (not yet implemented)
int  statHostCount = 0
int  statFileCount = 0
long statTotalSize = 0

// Check that the working volumes have sufficient available space
// (defaults to 1000MB if not specified in sasev.properties)

if (!checkLZSpace()) {
     msg.error "Required minimum LandingZone working disk space not met. Exiting."
     return exitPgm(RC_GENERAL_ERROR)
}


try {
	initializeMetrics metricsFile

	Validator val = new Validator(this, [silent:true], verbose? log:null)
	ValidateResults vResults = val.validate(2)
	if (!vResults) {
	    msg.error "Service Architecture validation failed. Errors are:"
	    vResults.errors.each { msg.error "\t$it" }
	    msg.error "Service Architecture Log Centralization failed with RC $RC_GENERAL_ERROR"
	    return exitPgm(RC_GENERAL_ERROR)
	}

	String centralLandingZone = sysProps['sasev.landingZoneDir']

	if (sysProps["logCentralize.enabled"]==null || logCentralizeEnabled) {
		AntBuilder ant = new AntBuilder()
		ant.project.getBuildListeners()[0].setMessageOutputLevel(verbose? 2 : 0)

		// Iterate over each of the registered log collectors
		hq.getResources("SAS System Info Log Collection").each { logCollector ->

			logCollector = hq.api.resourceApi.getResource(logCollector.id, true, false).resource
			def platform = hq.api.resourceApi.getParent(
				hq.api.resourceApi.getParent(logCollector).resource
			).resource

			String host = findValue(platform.resourceInfo, 'fqdn')
			String remoteLandingZone = findValue(logCollector.resourceProperty, 'landing.zone')

			// for testing purposes to collect from only a single host
			String customHost = sysProps["custom.logCollector.host"] ?: ""
			if (customHost) {
				msg.info "Custom Host Override set to [${customHost}]"
				if (customHost != host) {
					msg.info "Custom Host [${customHost}] Does not match host [${host}]  skipping"
					return // from closure
				}
			}

			msg.info "Gathering logs from $host..."
			msg.debug "  remoteLandingZone  = $remoteLandingZone"
			msg.debug "  centralLandingZone = $centralLandingZone"

			canonicalFile(centralLandingZone).mkdirs()

			if (net.isLocal(host)) {
				// Local copy
				(verbose? msg:log).info "Gathering local logs:"
				(verbose? msg:log).info "  Source     : $remoteLandingZone"
				(verbose? msg:log).info "  Destination: $centralLandingZone"

				ant.move(todir:centralLandingZone, overwrite:true, failonerror:false, preservelastmodified:true) {
					fileset(dir:remoteLandingZone) {
						include(name:'*.*')
						exclude(name:'*.json')
						type(type:'file')
					}
				}
			} else {
				// Remote copy
				def ctrl = da.control(host)
				boolean daAlreadyRunning = ctrl.hostAvailable
				if (!daAlreadyRunning) {
					msg.info "Starting Deployment Agent on $host..."
					if (da.start(host)) {
						// Wait a bit for the Deployment Agent to initialize
						Thread.sleep(3000)
					} else {
						msg.warn "\tUnable to start Deployment Agent on $host.  Skipping."
						return // from closure to next collector
					}
				}

				(verbose? msg:log).info "Gathering remote logs:"
				(verbose? msg:log).info "  Source     : $remoteLandingZone"
				(verbose? msg:log).info "  Destination: $centralLandingZone"
				def daResults = ctrl.pullFolderFiles(remoteLandingZone, centralLandingZone)
				if (!daResults.status == 0) {
					term.warn "Remote copy unsuccessful for $host"
					log.warn "Remote copy unsuccessful for $host: ${daResults.dump()}"
				} else {
					(verbose? msg:log).info "Removing remote files..."
					// Clear out local landing zone if successful
					String[] cmd = null
					// Only Windows will have backslashes in the path
					if (remoteLandingZone.contains('\\')) {
						cmd = ['cmd', '/c', 'del', '/q', "$remoteLandingZone\\*.*"]
					} else {
						cmd = ['/bin/sh', '-c', "rm -rf $remoteLandingZone/*"]
					}
					daResults = ctrl.doExec(cmd)
					if (!daResults.status == 0) {
						term.warn "Remote Landing Zone cleanup unsuccessful for $host"
						log.warn "Remote Landing Zone cleanup unsuccessful for $host: ${daResults.dump()}"
					}
				}
				ctrl.disconnect()

				if (!daAlreadyRunning) {
					msg.info "Stopping Deployment Agent on $host..."
					da.stop(host)
					Thread.sleep(2000)
				}
			}
		}
	} else {
		msg.info "Log centralization disabled.  Nothing to do."
	}

	long duration = System.currentTimeMillis() - masterStartTime
	msg.info "Log Centralization completed in - ${duration/1000} sec"

	// Writing info event record
	events.info "Log Centralization completed in - ${duration/1000} sec"

	completeMetrics metricsFile, RC_SUCCESS
	return exitPgm(RC_SUCCESS)
} catch (Exception e) {
    term.error "There was a problem centralizing log files: ${e.message}"
    log.error "There was a problem centralizing log files", e
    errorCount++
	completeMetrics metricsFile, RC_GENERAL_ERROR
	return exitPgm(RC_GENERAL_ERROR)
}

def checkMinSpace(String myprop, long defaultMinMB = 1000) {

// 'myprop' is a 'sysProps' property string representing a directory path.
// This routine will append ".MinSizeMB" to the supplied 'sysProps' property string
// and will use that to fetch the associated minimum required disk space for that path (in MB),
// comparing the usable space at that path to the supplied (or default) minimum amount.
// Returns 'false' if minimum value not met.

    long mySpaceMB

    String myminprop = myprop + '.MinSizeMB'

    long myminMB = Long.valueOf(sysProps[myminprop] ? sysProps[myminprop] : defaultMinMB)
    File myFile = canonicalFile(sysProps[myprop])

    try {
        mySpaceMB = myFile.getUsableSpace() / (1000 * 1000)
    } catch (Exception e) {
        term.error "Could not fetch usable space for $myFile: $e"
        log.error "Could not fetch usable space for $myFile", e
    }

    msg.debug "checkMinSpace: myprop is $myprop"
    msg.debug "checkMinSpace: Usable space in MB is $mySpaceMB, minimum limit is $myminMB\n"

    if (mySpaceMB < myminMB) {
        msg.error "Not enough disk space for centralization:"
        msg.error "   $myprop = $myFile"
        msg.error "   Usable space in MB is $mySpaceMB, required is $myminMB"
        return false
    }
    return true
}

def checkLZSpace() {
    space_OK = true

    if (!checkMinSpace('sasev.landingZoneDir')) {
        space_OK = false
        errorCount++
    }
    return space_OK
}

// Finds a value in a list of objects that have getKey()/getValue()
String findValue(List list, String key) {
	String result = null
	if (list && key) {
		list.each {
			if (it.key == key) {
				result = it.value
			}
		}
	}

	return result
}

def exitPgm(int rc) {
    if (exitJVM) {
        System.exit rc
    } else {
        return rc
    }
}
