/*****************************************************************************
 * Copyright © 2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
 ****************************************************************************/

// NOTE: Requires Service Architecture runtime and will not function standalone

@BaseScript(MasterETLBaseScript)
import groovy.transform.BaseScript
import java.io.File

def pgmName = this.class.name
scratchDir = null
altLogDir = canonicalFile(sysProps["emi.log.dir"], "kits/emi-framework/Logs/")

// Set up expected CLI syntax
def cli = new CliBuilder(
    usage:  "master_acm_etl${scriptExtension} [-xvdh] [--evdebug <0|1|2|3>]",
    header: "Executes the ETL process for the ACM component\n" +
            "Options:",
    footer: "Exit status:\n" +
            "$RC_SUCCESS  if OK\n" +
            "$RC_WARNING  if minor issues are encountered\n" +
            "$RC_GENERAL_ERROR  if an error is encountered\n" +
            "$RC_INVALID_USAGE for invalid usage"
)
cli.with {
    x(longOpt: 'noexit', 'Do not exit the JVM on completion', required: false)
    _(longOpt: 'evdebug', 'Set debug level for SAS invocations', args:1, type: int, argName: '0-3', required: false)
    v(longOpt:'verbose', 'Enable verbose output', required: false)
    d(longOpt:'debug', 'Enable debug output', required: false)
    h(longOpt:'help', 'Show usage information', required: false)
}

// Parse given arguments
OptionAccessor opts = cli.parse(args)
if (!opts) {
    return exitETL(RC_INVALID_USAGE)
}
if (opts.h) {
    cli.usage()
    return exitETL(RC_SUCCESS)
}
if (opts.v) {
    verbose = true
}
if (opts.d) {
    setLogLevel LogLevel.DEBUG
    evDebug = 2
}
if (opts.x) {
    exitJVM = false
}
if (opts.evdebug) {
    evDebug = opts.evdebug as int
    if (!evDebug in [0-3]) {
        msg.error "Invalid value for evdebug option ($evDebug).  Value should be 0-3"
        return exitETL(RC_INVALID_USAGE)
    }
}

// Property setting to enable and trigger the runReports capability ( only the appropriate report
// report bundle is selected) via the acm etl. run results method
boolean runReports = false
if (sysProps['custom.etl.runReports'] ==~ /(?i)true/) {
   runReports = true
}

if (!frameworkInitialized) {
    msg.error "The ACM ETL cannot run since Service Architecture is not initialized"
    errorCount++
    return exitETL(RC_GENERAL_ERROR)
}

msg.info "Service Architecture ACM ETL Process - ${new Date()}"

if (isRunning("${this.class.name}" + ".groovy")) {
    msg.error "The ACM ETL is already running. The existing process must complete before continuing."
    errorCount++
    return RC_GENERAL_ERROR
}

// Run Validator level 2
Validator v = new Validator(this, [verbose:verbose], log)
ValidateResults vResults = v.validate(2)
if (!vResults) {
    msg.error "Service Architecture validation failed. ${pgmName} cannot continue. Errors are:"
    vResults.errors.each { msg.error "\t$it" }
    msg.error "Stopping execution"
    return exitETL(RC_GENERAL_ERROR)
}

if (!acmEnabled) {
    msg.error "ACM is not enabled.  The ACM ETL cannot be run."
    errorCount++
    return exitETL(RC_WARNING)
}

runID = _LOG_TIMESTAMP

// Federated Data Mart export
boolean exportData = sysProps['datamart.export.withETL'] ==~ /(?i)true/

// Copy data to VA Autoload dropzone
boolean copy2VA = sysProps['vafeed.enabled'] ==~ /(?i)true/

String deployid = sysProps["sasev.deployment.id"] ?: net.fqdn.replaceAll(/\./,'_')
String deploymentLabel= sysProps["sas.deployment.label"] ?: "No Label"

// Check that the working volumes have sufficient available space
// (defaults to 1000MB if not specified in sasev.properties)

if (!checkETLSpace()) {
     msg.error "Required minimum ETL working disk space not met. Exiting."
     return exitETL(RC_GENERAL_ERROR)
}

// Create scratchDir if needed
evdmTempDir = canonicalFile(datamartDir, "evdmTemp")
scratchDir = canonicalFile(evdmTempDir, "scratch_${runID}")

if (!scratchDir.mkdir()) {
     msg.error "Cannot create [$scratchDir]. Exiting."
     return exitETL(RC_GENERAL_ERROR)
}

commonSASOptions.env = acmEnv + [EVDEBUG:"$evDebug"] + [RUNID:"$runID"] + [DEPLOYID:"$deployid"]
commonSASOptions.verbose = verbose
commonSASOptions.args = COMMON_SAS_ARGS
commonSASOptions.logDir = logDir.absolutePath
commonSASOptions.autoexec = canonicalPath(datamartDir, "EVDMAUTOEXEC_BATCH.sas")

// Make sure the landing zone directory for altlogs exists
altLogDir.mkdirs()

// Main body of ETL

try {
	metricsFile = canonicalFile(metricsDir, "acm_etl.metrics")
    initializeMetrics metricsFile

    Map myOpts = commonSASOptions + ["logPrefix":"acm"]
    dbMirrorData(metricsFile, myOpts)
    inventoryUpdate(metricsFile, myOpts)

    metricData()

    availabilityData()

    eventData()

    evdmUpdate()

    if (exportData) {
       msg.info "Exporting Data."
       exportSomeData()
    }

    evdmMetrics()

    runSASReports(runReports)

    if (copy2VA) {
       vaCopy()
    }
    
    copyLogs()

} catch (Exception e) {
    term.error "There was a problem executing the ACM ETL: $e"
    log.error "There was a problem executing the ACM ETL", e
    errorCount++
}

int rc = errorCount==0? RC_SUCCESS : RC_GENERAL_ERROR
completeMetrics metricsFile, rc


events.info "$pgmName ran"

long duration = System.currentTimeMillis() - masterStartTime
term.info "Check $logFile for messages"
msg.info "ACM Master ETL Script completed in $duration ms at ${new Date()} with $errorCount error(s)"

return exitETL(rc)
// =====================================================================
// End of Script
// =====================================================================

// =====================================================================
// Step Implementations
// =====================================================================

def checkMinSpace(String myprop, long defaultMinMB = 1000) {

// 'myprop' is a 'sysProps' property string representing a directory path.
// This routine will append ".MinSizeMB" to the supplied 'sysProps' property string
// and will use that to fetch the associated minimum required disk space for that path (in MB),
// comparing the usable space at that path to the supplied (or default) minimum amount.
// Returns 'false' if minimum value not met.

    long mySpaceMB

    String myminprop = myprop + '.MinSizeMB'

    long myminMB = Long.valueOf(sysProps[myminprop] ? sysProps[myminprop] : defaultMinMB)
    File myFile = canonicalFile(sysProps[myprop])

    try {
        mySpaceMB = myFile.getUsableSpace() / (1000 * 1000)
    } catch (Exception e) {
        term.error "Could not fetch usable space for $myFile: $e"
        log.error "Could not fetch usable space for $myFile", e
    }

    msg.debug "checkMinSpace: myprop is $myprop"
    msg.debug "checkMinSpace: Usable space in MB is $mySpaceMB, minimum limit is $myminMB\n"

    if (mySpaceMB < myminMB) {
        msg.error "Not enough disk space for ETL:"
        msg.error "   $myprop = $myFile"
        msg.error "   Usable space in MB is $mySpaceMB, required is $myminMB"
        return false
    }
    return true
}

def checkETLSpace() {
    space_OK = true

    if (!checkMinSpace('sasev.datamartDir')) {
        space_OK = false
        errorCount++
    }
    return space_OK
}

// Extract/load the SASEV Datamart Metric Data
int metricData() {
    long start = System.currentTimeMillis()
    msg.info "Starting metricData"
    int numErrors = 0

    File sasProg = canonicalFile(jobsDir, "getMetricData.sas")
    Map myOpts = [:] << commonSASOptions
    myOpts["args"] += [
    	"-altlog", canonicalPath(altLogDir, "${sasProg.name-'.sas'}_${_LOG_TIMESTAMP}.sas.log")
    	]

    int rc = sas.run myOpts, sasProg

    long runTime = (System.currentTimeMillis() - start)/1000
    writeMetric(
        "metricExtractProcessingTime",
        "$runTime",
        "Elapsed execution time of metric extract script",
        metricsFile
    )
    writeMetric(
        "metricExtractReturnCode",
        "$rc",
        "Return code from  metrc extract script",
        metricsFile
    )
    if (rc == 1) {
        term.warn "${sasProg.name} completed with WARNINGS"
    } else if (rc > 1) {
        term.error "${sasProg.name} completed with ERRORS"
        numErrors++
    }

    msg.info "Completed metricData in $runTime seconds with $numErrors error(s)"
    errorCount += numErrors
    return numErrors
}

// Extract/load the SASEV Datamart Availability Data
int availabilityData() {
    long start = System.currentTimeMillis()
    msg.info "Starting availabilityData"
    int numErrors = 0

    File sasProg = canonicalFile(jobsDir, "getAvailData.sas")
    Map myOpts = [:] << commonSASOptions
    myOpts["args"] += [
    	"-altlog", canonicalPath(altLogDir, "${sasProg.name-'.sas'}_${_LOG_TIMESTAMP}.sas.log")
    	]

    int rc = sas.run myOpts, sasProg

    long runTime = (System.currentTimeMillis() - start)/1000
    writeMetric(
        "availabilityExtractProcessingTime",
        "$runTime",
        "Elapsed execution time of availability extract script",
        metricsFile
    )
    writeMetric(
        "availabilityExtractReturnCode",
        "$rc",
        "Return code from availability extract script",
        metricsFile
    )
    if (rc == 1) {
        term.warn "${sasProg.name} completed with WARNINGS"
    } else if (rc > 1) {
        term.error "${sasProg.name} completed with ERRORS"
        numErrors++
    }

    msg.info "Completed availabilityData in $runTime seconds with $numErrors error(s)"
    errorCount += numErrors
    return numErrors
}

// Extract/load the SASEV Datamart Event Data
int eventData() {
    long start = System.currentTimeMillis()
    msg.info "Starting eventData"
    int numErrors = 0

    File sasProg = canonicalFile(jobsDir, "getEventData.sas")
    Map myOpts = [:] << commonSASOptions
    myOpts["args"] += [
    	"-altlog", canonicalPath(altLogDir, "${sasProg.name-'.sas'}_${_LOG_TIMESTAMP}.sas.log")
    	]

    int rc = sas.run myOpts, sasProg

    long runTime = (System.currentTimeMillis() - start)/1000
    writeMetric(
        "eventExtractProcessingTime",
        "$runTime",
        "Elapsed execution time of event extract script",
        metricsFile
    )
    writeMetric(
        "eventExtractReturnCode",
        "$rc",
        "Return code from event extract script",
        metricsFile
    )
    if (rc == 1) {
        term.warn "${sasProg.name} completed with WARNINGS"
    } else if (rc > 1) {
        term.error "${sasProg.name} completed with ERRORS"
        numErrors++
    }

    msg.info "Completed eventData in $runTime seconds with $numErrors error(s)"
    errorCount += numErrors
    return numErrors
}

// Update the SASEV Datamart Reporting Tables
int evdmUpdate() {
    long start = System.currentTimeMillis()
    msg.info "Starting evdmUpdate"
    int numErrors = 0

    File sasProg = canonicalFile(jobsDir, "updateEVDM.sas")
    Map myOpts = [:] << commonSASOptions
    myOpts["args"] += [
    	"-altlog", canonicalPath(altLogDir, "${sasProg.name-'.sas'}_${_LOG_TIMESTAMP}.sas.log")
    	]

    int rc = sas.run myOpts, sasProg

    long runTime = (System.currentTimeMillis() - start)/1000
    writeMetric(
        "reportTablesProcessingTime",
        "$runTime",
        "Elapsed execution time of the reporting table(s) update script",
        metricsFile
    )
    writeMetric(
        "reportTablesReturnCode",
        "$rc",
        "The return code from the reporting table(s) update script",
        metricsFile
    )
    if (rc == 1) {
        term.warn "${sasProg.name} completed with WARNINGS"
    } else if (rc > 1) {
        term.error "${sasProg.name} completed with WARNINGS"
        numErrors++
    }

    msg.info "Completed evdmUpdate in $runTime seconds with $numErrors error(s)"
    errorCount += numErrors
    return numErrors
}

// Handle Exporting ACM Data if necessary
int exportSomeData() {

    argsList = ['-x']
    if (verbose) argsList << '-v'
    if (msg.debugEnabled) argsList << '-d'
    argsList << '-acmlibmode'
    argsList << 'DELTAS'
// export of APM data currently not supported
// therefore following args should NOT be passed
//    argsList << '-apmlibmode'
//    argsList << 'NONE'
    argsList << '-runID'
    argsList << runID

    msg.info "Calling Data Mart Export script with options ${argsList}"
    rc = runGroovy canonicalPath(scriptDir,'exportDataMart.groovy'), argsList

    return rc

}

// Calculate overall SASEV Datamart Metrics
int evdmMetrics() {
    long start = System.currentTimeMillis()
    msg.info "Starting evdmMetrics"
    int numErrors = 0

    File sasProg = canonicalFile(jobsDir, "sasevdm_metrics.sas")
    File mFile = canonicalFile(metricsDir, "sasevdm.metrics")
    initializeMetrics mFile
    Map myOpts = [:] << commonSASOptions
    Map myEnv = [METRICSFILE:mFile.absolutePath] + myOpts["env"]
    myOpts += [env:myEnv]
    myOpts["args"] += [
    	"-altlog", canonicalPath(altLogDir, "${sasProg.name-'.sas'}_${_LOG_TIMESTAMP}.sas.log")
    	]

    int rc = sas.run myOpts, sasProg

    long runTime = (System.currentTimeMillis() - start)/1000
    if (rc == 1) {
        term.warn "${sasProg.name} completed with WARNINGS"
    } else if (rc > 1) {
        term.error "${sasProg.name} completed with ERRORS"
        numErrors++
    }

    msg.info "Completed evdmMetrics in $runTime seconds with $numErrors error(s)"
    errorCount += numErrors
    return numErrors
}

def runSASReports(boolean runReports) {
    msg.info "Starting runSASReports..."
    long start = System.currentTimeMillis()

    if (runReports) {

        // run groovy program
        msg.info "Running report bundles ACM"

        // call runReports need to support Audit, ARM, ACM, Kits
        List pArgs = ['--rptBundle','ACM']
        if (verbose) pArgs << '-v'
        if (msg.debugEnabled) pArgs << '-d'
        verbose? msg.info ("running batch report bundle ${pArgs}") : ""
        int rc = runGroovy(canonicalPath(sysProps['emi.framework.scripts.dir'],'runReports.groovy'), pArgs)
        if (rc > 1) {
            term.error "$sasProg.name completed with ERRORS rc [$rc]"
            errorCount++
        }
    }

    long duration = System.currentTimeMillis() - start
    (verbose? msg:log).info "runSASReports completed in $duration ms"
}

def vaCopy() {
    msg.info "Starting vaCopy..."
    long start = System.currentTimeMillis()

    File vacMetricsFile = canonicalFile(metricsDir, "acm_vacopy.metrics")
    initializeMetrics vacMetricsFile
    Map myOpts = [:] << commonSASOptions
    myOpts['env'] = acmEnv + [METRICSFILE:vacMetricsFile.absolutePath] + [DATA2COPY2VA:"ACM2VA"]   

    myOpts['autoexec'] = canonicalPath(datamartDir, "EVDMAUTOEXEC_BATCH.sas")
    myOpts['logPrefix'] = "acm"
    copy2VA(vacMetricsFile, myOpts)

    long duration = System.currentTimeMillis() - start
    (verbose? msg:log).info "vaCopy completed in $duration ms"
}

// Copy logs to APM landing zone
int copyLogs() {
    long start = System.currentTimeMillis()
    msg.info "Starting copyLogs"

    int numErrors = 0

    try {
        File logDest = canonicalFile(sysProps['sasev.landingZoneDir'], "kits/sasev/Logs")
        logDest.mkdirs()

        AntBuilder ant = new AntBuilder()
        ant.project.buildListeners[0].messageOutputLevel = 0
        ant.copy(todir: logDest.absolutePath) {
            fileset(dir : logDir.absolutePath) {
                include(name:"**/*${_LOG_TIMESTAMP}.sas.log")
            }
        }
    } catch (Exception e) {
        term.warn "Unable to copy files to landing zone: $e"
        log.warn "Unable to copy files to landing zone", e
        numErrors++
    }
    long runTime = (System.currentTimeMillis() - start)/1000

    msg.info "Completed copyLogs in $runTime seconds with $numErrors error(s)"
    errorCount += numErrors
    return numErrors
}

// =====================================================================
// Helper Methods
// =====================================================================
int exitETL(int rc) {
    if (scratchDir)  {
        if (scratchDir.exists())  {
            if (scratchDir.deleteDir()) {
                msg.info "Deleting scratch area"
            } else {
                msg.error "Cannot delete [$scratchDir]"
                rc = RC_GENERAL_ERROR
            }
        }
    }

    log.info "ACM ETL completed with return code $rc"
    if (exitJVM) {
        System.exit rc
    } else {
        return rc
    }
}
